"""
Database utilities for SQLAlchemy.
"""

import logging
import os
import re

from sqlalchemy import text
from sqlalchemy.engine import make_url
from sqlalchemy.ext.asyncio import create_async_engine

logger = logging.getLogger(__name__)

re_valid_db_name = re.compile(r"^[a-zA-z0-9_\-]+$")


def get_async_database_url(db_url):
    """
    Return a parsed database URL, ensuring the async version is being used.
    """

    parsed_url = make_url(db_url)

    if parsed_url.drivername == "postgresql+asyncpg":
        # Already using the async version
        return parsed_url

    if parsed_url.drivername == "postgresql":
        return parsed_url.set(drivername="postgresql+asyncpg")

    raise ValueError(f"Unsupported driver: {parsed_url.drivername}")


def create_engine(database_url, **kwargs):
    db_url = get_async_database_url(database_url)

    if str(db_url) != str(database_url):
        logger.warning(
            "Database URL %s does not support asyncio, changing it to %s",
            repr(make_url(database_url)),
            repr(make_url(db_url)),
        )

    return create_async_engine(db_url, **kwargs)


def create_test_database_name():
    return "test_database_{}".format(os.urandom(6).hex())


async def create_database(db_url, database_name):
    if not validate_database_name(database_name):
        raise ValueError(f"Invalid database name: {database_name}")

    engine = create_engine(db_url, isolation_level="AUTOCOMMIT")

    async with engine.connect() as conn:
        c = await conn.execute(
            text(f"SELECT 1 FROM pg_database WHERE datname='{database_name}'")
        )
        database_exists = c.scalar() == 1

    if database_exists:
        await drop_database(db_url, database_name)

    async with engine.connect() as conn:
        await conn.execute(text(f'CREATE DATABASE "{database_name}" ENCODING "utf8"'))
    await engine.dispose()


async def drop_database(db_url, database_name):
    if not validate_database_name(database_name):
        raise ValueError(f"Invalid database name: {database_name}")

    engine = create_engine(db_url, isolation_level="AUTOCOMMIT")

    async with engine.connect() as conn:
        disc_users = """
        SELECT pg_terminate_backend(pg_stat_activity.%(pid_column)s)
        FROM pg_stat_activity
        WHERE pg_stat_activity.datname = '%(database)s'
          AND %(pid_column)s <> pg_backend_pid();
        """ % {
            "pid_column": "pid",
            "database": database_name,
        }
        await conn.execute(text(disc_users))

        await conn.execute(text(f'DROP DATABASE "{database_name}"'))


def validate_database_name(name):
    return bool(re_valid_db_name.match(name))



from __future__ import annotations

import abc
import typing

if typing.TYPE_CHECKING:
    from email.message import Message


class BaseMailer(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def send_message(self, msg: Message):
        pass



import smtplib

from .base import BaseMailer


class SMTPMailer(BaseMailer):
    def __init__(self, host, port=25, starttls=True, username=None, password=None):
        self._host = host
        self._port = port
        self._starttls = starttls
        self._username = username
        self._password = password

    def send_message(self, msg):
        with smtplib.SMTP(self._host, port=self._port) as smtp:
            if self._starttls:
                smtp.starttls()
            if self._username or self._password:
                smtp.login(self._username, self._password)
            smtp.send_message(msg)



import dataclasses
from dataclasses import dataclass

import pytest

from mowaki.config import create_config_from_env


class Test_dataclass_field_parsing:
    def test_field_with_no_default_is_required(self):
        @dataclass
        class SomeObject:
            somefield: str

        fields = dataclasses.fields(SomeObject)
        [field] = fields
        assert field.name == "somefield"
        assert field.default is dataclasses.MISSING
        assert field.default_factory is dataclasses.MISSING
        assert field.type is str


class Test_load_string_config:
    def test_required_string_is_loaded(self):
        @dataclass
        class MyConfig:
            mystring: str

        env = {"MYSTRING": "Some value"}
        result = create_config_from_env(MyConfig, env)

        assert isinstance(result, MyConfig)
        assert result.mystring == "Some value"

    def test_required_string_cannot_be_missing(self):
        @dataclass
        class MyConfig:
            mystring: str

        env = {}
        with pytest.raises(KeyError):
            create_config_from_env(MyConfig, env)

    def test_omitted_optional_string_uses_default(self):
        @dataclass
        class MyConfig:
            mystring: str = "Default value"

        result = create_config_from_env(MyConfig, {})

        assert isinstance(result, MyConfig)
        assert result.mystring == "Default value"

    def test_omitted_optional_string_uses_default_factory(self):
        @dataclass
        class MyConfig:
            mystring: str = dataclasses.field(default_factory=lambda: "Default value")

        result = create_config_from_env(MyConfig, {})

        assert isinstance(result, MyConfig)
        assert result.mystring == "Default value"


class Test_load_int_config:
    def test_int_value_is_loaded(self):
        @dataclass
        class MyConfig:
            myint: int

        env = {"MYINT": "32"}
        result = create_config_from_env(MyConfig, env)

        assert isinstance(result, MyConfig)
        assert result.myint == 32


class Test_load_bool_config:
    @pytest.mark.parametrize("raw,expected", [
        ("TRUE", True),
        ("FALSE", False),
        ("yes", True),
        ("no", False),
        ("1", True),
        ("0", False),
        ("", False),
    ])
    def test_bool_value_is_loaded(self, raw, expected):
        @dataclass
        class MyConfig:
            mybool: bool

        env = {"MYBOOL": raw}
        result = create_config_from_env(MyConfig, env)

        assert isinstance(result, MyConfig)
        assert result.mybool is expected



import base64
import hashlib
import hmac
import json
import os


class InvalidPassword(Exception):
    pass


SCRYPT_PARAMS = {
    "n": 2**14,
    "r": 2**4,
    "p": 1,
    "maxmem": 128 * (1024**2),  # 128 MiB
    "dklen": 64,
}


def create_encrypted_password(plaintext: str) -> str:
    salt = os.urandom(16)
    key_bytes = hashlib.scrypt(plaintext.encode("utf-8"), salt=salt, **SCRYPT_PARAMS)
    data = {
        "method": "scrypt",
        "salt": salt.hex(),
        "params": SCRYPT_PARAMS,
        "key": key_bytes.hex(),
    }
    return base64.encodebytes(json.dumps(data).encode("utf-8")).decode("utf-8")


def validate_encrypted_password(encrypted: str, plaintext: str):
    decoded = base64.decodebytes(encrypted.encode("utf-8"))
    data = json.loads(decoded)

    if data["method"] == "scrypt":
        key_bytes = bytes.fromhex(data["key"])
        salt = bytes.fromhex(data["salt"])
        params = data["params"]
        result = hashlib.scrypt(plaintext.encode("utf-8"), salt=salt, **params)
        if not hmac.compare_digest(key_bytes, result):
            raise InvalidPassword("The provided password doesn't match")
        return True

    raise ValueError(f"Unsupported password encryption method: {data['method']}")



from urllib.parse import urlparse

from .base import BaseMailer
from .dummy import DummyMailer
from .smtp import SMTPMailer


def get_mailer_from_url(url: str) -> BaseMailer:
    """Create an appropriate Mailer instance from a URL

    Supported schemes:

    - smtp://username:password@host:port

        Send emails via an SMTP server.

    - dummy://

        Used for testing. Sent emails are instead stored in memory for
        further analysis.
    """

    parsed = urlparse(url)

    if parsed.scheme == "dummy":
        return DummyMailer()

    if parsed.scheme == "smtp":
        return SMTPMailer(
            host=parsed.hostname,
            port=parsed.port,
            username=parsed.username,
            password=parsed.password,
            starttls=True,
        )

    raise ValueError(f"Unsupported Mailer URL: {url}")



"""
High level functions to compose emails.

See https://docs.python.org/3/library/email.examples.html
"""

from email.utils import make_msgid
from email.message import EmailMessage
from typing import Iterable


def compose_text_email(
    text: str, *, subject: str, sender: str, to: str | Iterable[str]
) -> EmailMessage:
    msg = EmailMessage()
    msg["Subject"] = subject
    msg["From"] = sender
    msg["To"] = to
    msg.set_content(text)
    return msg


def compose_html_email(
    html_content, *, subject: str, sender: str, to: str, text_content=None
) -> EmailMessage:
    msg = EmailMessage()
    msg["Subject"] = subject
    msg["From"] = sender
    msg["To"] = to
    msg.set_content(text_content)
    msg.add_alternative(html_content, subtype="html")
    return msg



import mimetypes
import os
from typing import Any, Dict, NamedTuple, Type  # noqa
from urllib.parse import quote, urlparse

from .base import BaseStorage
from .types import FileInfo, FileMetadata


class FileSystemStorage(BaseStorage):
    """Filesystem-based local storage

    Warning:
        This is meant to be used for development, not production use!
    """

    def __init__(self, root_path: str) -> None:
        self.root_path = root_path

    @classmethod
    def from_url(cls, url):  # type: (str) -> BaseStorage
        parsed = urlparse(url)
        if parsed.netloc:
            raise ValueError(
                'File path cannot contain a netloc. '
                'Make sure the URL contains three forward slashes, eg. '
                'file:///path/to/storage')
        return cls(parsed.path)

    def get_file(self, key: str) -> FileInfo:
        return FileInfo(metadata=self.get_file_meta(key),
                        content=self.get_file_content(key))

    def get_file_meta(self, key: str) -> FileMetadata:
        path = self._get_file_path(key)
        if not os.path.exists(path):
            return None
        content_type, content_encoding = mimetypes.guess_type(path)
        return FileMetadata(
            content_type=content_type)

    def get_file_content(self, key: str) -> bytes:
        path = self._get_file_path(key)
        with open(path, 'rb') as fp:
            return fp.read()

    def put_file(self, key: str, data: bytes, mime_type: str = None) -> None:
        path = self._get_file_path(key)
        with open(path, 'wb') as fp:
            fp.write(data)

    def _get_file_path(self, key: str) -> str:
        key = quote(key, safe='')
        return os.path.join(self.root_path, key)

    def get_file_url(self, key: str) -> str:
        full_path = self._get_file_path(key)
        return 'file://{}'.format(full_path)

    def file_exists(self, key: str) -> bool:
        return os.path.exists(self._get_file_path(key))

    def get_etag(self, key: str) -> str:
        return None



from datetime import datetime, timedelta

import jwt


class TokenMaker:

    """Convenience object for issuing and validating JWT tokens.

    Keeps track of (and validates) secret, algorithm, issuer, and
    audience for a particular JWT token type.
    """

    def __init__(
        self,
        secret,
        issuer=None,
        audience=None,
        validity=None,
        algorithm="HS256",
    ):
        self.secret = secret
        self.issuer = issuer
        self.audience = audience
        self.validity = validity
        self.algorithm = algorithm

    def issue(self, subject, validity=None, **kwargs):
        """Issue a new JWT token

        Args:

            subject (str):
                Token subject (eg. user ID)

            validity (datetime.timedelta):
                Token validity.

            exp (datetime.datetime):
                Token expiration date. Overrides validity.
        """

        claim = {
            "sub": subject,
        }

        if self.issuer is not None:
            claim["iss"] = self.issuer

        if self.audience is not None:
            claim["aud"] = self.audience

        if validity is None:
            validity = self.validity

        if validity is not None:
            claim["exp"] = datetime.utcnow() + timedelta(seconds=self.validity)

        claim.update(kwargs)

        return jwt.encode(claim, self.secret, algorithm=self.algorithm)

    def validate(self, token):
        """Validate a JWT token, returning the contained data"""

        return jwt.decode(
            token,
            self.secret,
            algorithms=[self.algorithm],
            issuer=self.issuer,
            audience=self.audience,
        )


def get_unverified_header(token):
    return jwt.get_unverified_header(token)


def get_unverified_token_data(token):
    return jwt.decode(token, options={"verify_signature": False})


def get_token_audience(token):
    return get_unverified_token_data(token)["aud"]



"""
Configuration management utilities
"""

import dataclasses
from typing import Type, TypeVar

AppConfigType = TypeVar("AppConfigType")
T = TypeVar("T")


def create_config_from_env(
    config_type: Type[AppConfigType],
    env: dict[str, str],
) -> AppConfigType:
    """
    Create an instance of AppConfig from environment variables.

    Args:

        config_type:
            Class to be instantiated with the configuration.
            Usually a dataclass (fields names are obtained by calling
            dataclasses.fields() on the type).

        env:
            Dict of environment variables to read. Keys and values are
            expected to be strings.

        alt_prefixes:
            List of alternate prefixes to search configuration
            variables.  Most useful during testing, to allow
            overriding configuration with TEST_* variables.

    Returns:

        A configured instance of ``config_type``
    """

    fields = dataclasses.fields(config_type)
    kwargs = {}

    for field in fields:
        env_key = field.name.upper()

        try:
            value = env[env_key]
        except KeyError:
            if _is_required_field(field):
                raise
            value = _get_field_default(field)

        kwargs[field.name] = _parse_value(value, field.type)

    return config_type(**kwargs)


def _is_required_field(field):
    """
    If the field doesn't have either a default value or default
    factory, then it must be supplied.
    """
    return (field.default is dataclasses.MISSING) and (
        field.default_factory is dataclasses.MISSING
    )


def _get_field_default(field):
    """Get the default value for a dataclass field"""

    if field.default is not dataclasses.MISSING:
        return field.default

    if field.default_factory is not dataclasses.MISSING:
        return field.default_factory()

    return None


def _parse_value(value: str, type_: Type[T]) -> T:
    if type_ is str:
        return value

    if type_ is int:
        return int(value)

    if type_ is float:
        return float(value)

    if type_ is bool:
        return _parse_bool(value)

    raise ValueError(f"Unsupported configuration variable type: {type_}")


def _parse_bool(value: str) -> bool:
    TRUTHY = {"true", "yes", "on", "1"}
    FALSEY = {"false", "no", "off", "0", ""}

    _value = value.lower()

    if _value in TRUTHY:
        return True

    if _value in FALSEY:
        return False

    raise ValueError(f"Unable to parse boolean value: {value}")



import os
from setuptools import setup, find_packages

version = '2.0'

here = os.path.dirname(__file__)

with open(os.path.join(here, 'README.rst')) as fp:
    longdesc = fp.read()

with open(os.path.join(here, 'CHANGELOG.rst')) as fp:
    longdesc += "\n\n" + fp.read()


setup(
    name='mowaki',
    version=version,
    packages=find_packages(),
    url='https://github.com/rshk/mowaki-py',

    license='BSD License',

    author='Samuele Santi',
    author_email='samuele.santi@reinventsoftware.io',
    description='MoWAKi - Python utilities',
    long_description=longdesc,

    install_requires=[],
    extras_require={
        'aws': ['boto3'],
        'auth': ['pyjwt'],
        'dev': [
            "pytest>=5.4",
            "pytest-cov",
        ],
    },
    # tests_require=tests_require,
    # test_suite='tests',

    classifiers=[
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',

        # 'Programming Language :: Python :: Implementation :: CPython',
        # 'Programming Language :: Python :: Implementation :: IronPython',
        # 'Programming Language :: Python :: Implementation :: Jython',
        # 'Programming Language :: Python :: Implementation :: PyPy',
        # 'Programming Language :: Python :: Implementation :: Stackless',
    ],
    # entry_points={
    #     'console_scripts': ['mowaki=mowaki.cli:main'],
    # },
    package_data={'': ['README.rst', 'CHANGELOG.rst']},
    include_package_data=True,
    zip_safe=False)



from contextlib import contextmanager
from contextvars import ContextVar

from .base import BaseMailer

_outbox = ContextVar("dummy_outbox")


class DummyMailer(BaseMailer):
    def __init__(self):
        pass

    def send_message(self, msg):
        try:
            outbox = _outbox.get()
        except LookupError:
            pass
        else:
            outbox.append(msg)


@contextmanager
def record_sent_emails():
    """
    Context manager to record emails sent by dummy mailers.

    Most useful for testing. Will intercept all mail sent by dummy
    mailers in the current thread / execution context.
    """
    outbox = []
    token = _outbox.set(outbox)
    yield outbox
    _outbox.reset(token)



import logging
import posixpath
from urllib.parse import urlparse

import boto3
from botocore.client import Config

from .base import BaseStorage
from .types import FileInfo, FileMetadata, PresignedPostInfo, S3Object

logger = logging.getLogger(__name__)


_config = None


def get_aws_session() -> boto3.Session:
    return boto3.Session()


def get_s3_public_file_url(region: str, bucket_name: str, path: str) -> str:
    return 'https://{name}.s3-{region}.amazonaws.com/{path}'.format(
        region=region, name=bucket_name, path=path)


class S3Storage(BaseStorage):
    """Amazon S3 backed storage

    This is the recommended choice for production.
    """

    def __init__(self, bucket_name: str, key_prefix: str = None) -> None:
        self.bucket_name = bucket_name
        self.key_prefix = key_prefix or ''

    @classmethod
    def from_url(cls, url):  # type: (str) -> BaseStorage
        parsed = urlparse(url)
        if not parsed.netloc:
            raise ValueError(
                'S3 bucket name missing. '
                'Make sure the URL is in the form s3://bucket-name/key-prefix')
        return cls(parsed.netloc, key_prefix=parsed.path.lstrip('/'))

    def get_file(self, key: str) -> FileInfo:
        full_key = self._get_full_key(key)
        data = self._get_bucket().Object(full_key).get()
        return FileInfo(
            metadata=FileMetadata(
                content_type=data['ContentType']),
            content=data['Body'])

    def get_file_meta(self, key: str) -> FileMetadata:
        return self.get_file(key).metadata

    def get_file_content(self, key: str) -> bytes:
        return self.get_file(key).content

    def get_presigned_post(self, key: str, content_type: str) \
            -> PresignedPostInfo:
        full_key = self._get_full_key(key)

        # s3 = get_aws_session().resource('s3')
        aws = get_aws_session()
        s3 = aws.client('s3', config=Config(signature_version='s3v4'))

        presigned_post = s3.generate_presigned_post(
            Bucket=self.bucket_name,
            Key=full_key,
            Fields={"Content-Type": content_type},
            Conditions=[
                # TODO: limit file size herx
                {"Content-Type": content_type}
            ],
            ExpiresIn=3600)
        return PresignedPostInfo(
            url=presigned_post['url'],
            fields=presigned_post['fields'])

    def _get_bucket(self):
        aws = get_aws_session()
        s3 = aws.resource('s3')
        return s3.Bucket(self.bucket_name)

    def _get_full_key(self, key: str) -> str:
        return posixpath.join(self.key_prefix, key)

    def _get_object(self, key: str) -> S3Object:
        full_key = self._get_full_key(key)
        return self._get_bucket().Object(full_key)

    def put_file(self, key: str, data: bytes, mime_type: str = None) -> None:
        self._get_object(key).put(Body=data, ContentType=mime_type)

    def get_file_url(self, key: str) -> str:
        full_key = self._get_full_key(key)
        return ('s3://{bucket}/{key}'
                .format(bucket=self.bucket_name, key=full_key))

    def file_exists(self, key: str) -> bool:
        from botocore.exceptions import ClientError

        aws = get_aws_session()
        s3_client = aws.client('s3')
        try:
            s3_client.head_object(
                Bucket=self.bucket_name,
                Key=self._get_full_key(key))
        except ClientError as e:
            if e.response['Error']['Code'] == "404":
                return False
            raise
        return True

    def get_etag(self, key: str) -> str:
        return self._get_object(key).e_tag



from contextlib import contextmanager
from contextvars import ContextVar
from typing import Any


@contextmanager
def contextvar_contextmanager(ctxvar: ContextVar, value: Any):
    """
    ContextVar context manager.

    Sets a ContextVar to a certain value and resets it at the end of
    the execution block, irrespective of whether an exception was
    raised or not.
    """

    token = ctxvar.set(value)
    try:
        yield
    finally:
        ctxvar.reset(token)
